#include "Encoder.h"
#include "../Char.h"
#include "../String/StringBuilder.h"
#include "../String/Encoding.h"
namespace Themis {

CConfigBinEncoder::TData CConfigBinEncoder::Encode(TValueParam t) {
    StringBuilder builder;
    const uint8_t* pByte = &t[0];
    for (size_t i = 0, n = t.size(); i < n; i++) {
        builder.Append(L"0123456789abcdef"[(pByte[i] >> 4) & 0x0F]);
        builder.Append(L"0123456789abcdef"[pByte[i] & 0x0F]);
    }
    return builder.ToString();
}

CConfigBinEncoder::TValue CConfigBinEncoder::Decode(TDataParam t) {
    vector<uint8_t> vct;
    uint8_t abyTbl[UCHAR_MAX] = {0};
    int c;
    for (c = L'0'; c <= L'9'; c++) abyTbl[c] = static_cast<uint8_t>(c - L'0');
    for (c = L'a'; c <= L'f'; c++) abyTbl[c] = static_cast<uint8_t>(c - L'a' + 0x0a);
    for (c = L'A'; c <= L'F'; c++) abyTbl[c] = static_cast<uint8_t>(c - L'A' + 0x0A);
    for (int i = 0, nSize = t.GetLength() / 2; i < nSize; i++) {
        vct.push_back(static_cast<uint8_t>((abyTbl[t[i * 2]] << 4) | abyTbl[t[i * 2 + 1]]));
    }
    return vct;
}

CConfigStrEncoder::TData CConfigStrEncoder::Encode(TValueParam t) {
    StringBuilder builder;
    for (int i = 0, nLen = t.GetLength(); i < nLen; ) {
        wchar_t c = t[i++];
        if      (c == L'\\') { builder.Append(L"\\\\"); }
        else if (c == L'\r') { builder.Append(L"\\r"); }
        else if (c == L'\n') { builder.Append(L"\\n"); }
        else if (c == L'\0') { builder.Append(L"\\0"); }
        else {
            builder.Append(c);
        }
    }
    return builder.ToString();
}

CConfigStrEncoder::TValue CConfigStrEncoder::Decode(TDataParam t) {
    StringBuilder builder;
    for (int i = 0, nLen = t.GetLength(); i < nLen; ) {
        wchar_t c = t[i++];
        if (c == L'\\') {
            wchar_t c2 = t[i++];
            if      (c2 == L'\\') { builder.Append(L'\\'); }
            else if (c2 == L'r')  { builder.Append(L'\r'); }
            else if (c2 == L'n')  { builder.Append(L'\n'); }
            else if (c2 == L'0')  { builder.Append(L'\0'); }
            else                  { builder.Append(c); i--; }  //͖
        } else {
            builder.Append(c);
        }
    }
    return builder.ToString();
}

CConfigPointEncoder::TData CConfigPointEncoder::Encode(TValueParam t) {
    return string::Format(L"%d, %d", t.x, t.y);
}

CConfigPointEncoder::TValue CConfigPointEncoder::Decode(TDataParam t) {
    int x, y;
    if (sscanf(Encoding::GetDefault().GetBytes(t).c_str(), "%d, %d", &x, &y) == 2) {
        return CPoint(x, y);
    }
    throw RuntimeException(L"RtBOf[^(Wl)̓ǂݍ݂Ɏs܂");
}

CConfigSizeEncoder::TData CConfigSizeEncoder::Encode(TValueParam t) {
    return string::Format(L"%d, %d", t.cx, t.cy);
}

CConfigSizeEncoder::TValue CConfigSizeEncoder::Decode(TDataParam t) {
    int x, y;
    if (sscanf(Encoding::GetDefault().GetBytes(t).c_str(), "%d, %d", &x, &y) == 2) {
        return CPoint(x, y);
    }
    throw RuntimeException(L"RtBOf[^(cl)̓ǂݍ݂Ɏs܂");
}

CConfigRectEncoder::TData CConfigRectEncoder::Encode(TValueParam t) {
    return string::Format(L"%d, %d", t.left, t.top, t.GetWidth(), t.GetHeight());
}

CConfigRectEncoder::TValue CConfigRectEncoder::Decode(TDataParam t) {
    int x, y, z, a;
    if (sscanf(Encoding::GetDefault().GetBytes(t).c_str(), "%d, %d, %d, %d", &x, &y, &z, &a) == 4) {
        return CRect(x, y, z, a);
    }
    throw RuntimeException(L"RtBOf[^(`Wl)̓ǂݍ݂Ɏs܂");
}

CConfigColorEncoder::TData CConfigColorEncoder::Encode(TValueParam t) {
    return string::Format(L"%d, %d, %d", GetRValue(t), GetGValue(t), GetBValue(t));
}

CConfigColorEncoder::TValue CConfigColorEncoder::Decode(TDataParam t) {
    int x, y, z;
    if (sscanf(Encoding::GetDefault().GetBytes(t).c_str(), "%d, %d, %d", &x, &y, &z) == 3) {
        return RGB(x, y, z);
    }
    throw RuntimeException(L"RtBOf[^(RGBl)̓ǂݍ݂Ɏs܂");
}

}
