#include "../Windows/System.h"
#include "../ProcessExclusion.h"
#include "../File.h"
#include "Storage.h"
namespace Themis {

CConfigStorage::CConfigStorage() : m_bNoThrow(false) {
}

CConfigStorage::~CConfigStorage() {
}

CConfigStorage::TData CConfigStorage::Get(TNameParam oName) const {
    std::map<string, string>::const_iterator itr = m_oDataList.find(oName);
    if (itr != m_oDataList.end()) {
        return (*itr).second;
    }
    if (m_bNoThrow) {
        // ꉞO炢͏oƂB
        Themis_DebugLog(L"RtBOf[^ '" + oName + L"' " + L"̎擾Ɏs");
        return string();
    } else {
        throw RuntimeException(L"RtBOf[^ '" + oName + L"' " + L"̎擾Ɏs܂");
    }
}

void CConfigStorage::Set(TNameParam oName, TDataParam oValue) {
    Themis_DebugAssert(!oName.Contains(L" = "));
    m_oDataList[oName] = oValue;
}

void CConfigStorage::Clear() {
    m_oDataList.clear();
}

void CConfigStorage::Delete(TNameParam oName) {
    m_oDataList.erase(oName);
}

bool CConfigStorage::Exists(TNameParam oName) const {
    return (m_oDataList.find(oName) != m_oDataList.end());
}

void CConfigStorage::ChanegName(TNameParam oName1, TNameParam oName2) {
    if (Exists(oName1)) {
        Set(oName2, Get(oName1));
        Delete(oName1);
    }
}

vector<CConfigStorage::TName> CConfigStorage::GetNames() const {
    vector<TName> vct(m_oDataList.size());
    int i = 0;
    for (std::map<string, string>::const_iterator it = m_oDataList.begin(), itEnd = m_oDataList.end(); it != itEnd; ++it) {
        vct[i++] = (*it).first;
    }
    return vct;
}

vector<pair<CConfigStorage::TName, CConfigStorage::TData> > CConfigStorage::GetValues() const {
    vector<pair<TName, TData> > vct(m_oDataList.size());
    int i = 0;
    for (std::map<string, string>::const_iterator it = m_oDataList.begin(), itEnd = m_oDataList.end(); it != itEnd; ++it) {
        vct[i++] = *it;
    }
    return vct;
}

void CConfigStorage::SetValues(const vector<pair<TName, TData> >& vct) {
    for (size_t i = 0, n = vct.size(); i < n; i++) {
        Themis_DebugAssert(!vct[i].first.Contains(L" = "));
        m_oDataList[vct[i].first] = vct[i].second;
    }
}

CFileName CConfigStorage::GetDefaultFileName() const {
    return CFileName(CSystem::GetModuleFileName()).SetSuffix(L"cfg");
}

bool CConfigStorage::Save() const {
    return Save(GetDefaultFileName());
}

bool CConfigStorage::Load() {
    return Load(GetDefaultFileName());
}

bool CConfigStorage::Save(const CFileName& oFileName) const {
    try {
        CProcessExclusionLock lock(oFileName.Get());

        CFile(oFileName).Save(SaveToStr());
        return true;
    } catch (std::exception&) {
        Themis_DebugWarning(L"RtBOt@Cos");
    }
    return false;
}

string CConfigStorage::SaveToStr() const {
    string str;
    typedef std::map<string, string>::const_iterator TItr;
    for (TItr it = m_oDataList.begin(), itEnd = m_oDataList.end(); it != itEnd; it++) {
        str += (*it).first + L" = " + (*it).second + L"\r\n";
    }
    return str;
}

bool CConfigStorage::Load(const CFileName& oFileName) {
    try {
        CProcessExclusionLock lock(oFileName.Get());

        CFile file(oFileName);
        if (!file.Exists()) {
            return false;
        }
        LoadFromStr(file.LoadStr());
        return true;
    } catch (std::exception&) {
        Themis_DebugWarning(L"RtBOt@Cǂݍݎs");
    }
    return false;
}

bool CConfigStorage::LoadFromStr(const string& str) {
    vector<string> vct = str.Split(Array<string>(L"\r\n", L"\n", L"\r"));
    for (size_t i = 0, n = vct.size(); i < n; i++) {
        List<string> ss = vct[i].Split(Array<string>(L" ="), 2, StringSplitOptions::None);
        if (ss.GetCount() == 2) {
            Set(ss[0], ss[1].TrimStart());
        }
    }
    return true;
}

}
