#include "stdafx.h"
#include "File.h"
#include "FindFile.h"
#include "Windows/NativeString.h"
#include "Windows/PushErrorMode.h"
#include "String/Encoding.h"
namespace Themis {

namespace {
    static DWORD GetFileAttributesSafe(const string& str) {
        PushErrorMode errorMode(SEM_FAILCRITICALERRORS);
        return NativeString::IsUnicode() ? 
            ::GetFileAttributesW(str.ToCharPtr()) :
            ::GetFileAttributesA(Encoding::GetDefault().GetBytes(str).c_str());
    }
    static void SetFileAttributes(const string& str, DWORD dwAttributes) {
        BOOL result = NativeString::IsUnicode() ?
            ::SetFileAttributesW(str.ToCharPtr(), dwAttributes) :
            ::SetFileAttributesA(Encoding::GetDefault().GetBytes(str).c_str(), dwAttributes);
        Themis_DebugAssertApiLog(result != FALSE); result;
    }
}

bool CFile::Exists() {
    return (0xFFFFFFFF != GetFileAttributesSafe(GetName().Get()));
}

bool CFile::ExistsFile() {
    DWORD dwAttributes = GetFileAttributesSafe(GetName().Get());
    return (dwAttributes != 0xFFFFFFFF && (dwAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0);
}

bool CFile::ExistsDir() {
    DWORD dwAttributes = GetFileAttributesSafe(GetName().RemoveDirSeparator());
    return (dwAttributes != 0xFFFFFFFF && (dwAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0);
}

size_t CFile::GetSize() {
    try {
        if (ExistsDir()) {
            vector<CFileName> vct = CFile::FindFile(GetName(), -1);
            size_t nSum = 0;
            for (size_t i = 0, n = vct.size(); i < n; i++) {
                nSum += CFindFile(vct[i].Get()).GetSize();
            }
            return nSum;
        } else {
            return CFindFile(GetName().RemoveDirSeparator()).GetSize();
        }
    } catch (std::exception&) {
        Themis_DebugWarning(L"CFile::GetSize() : " + GetName().Get() + L"݂Ȃ");
        return 0;
    }
}

uint64_t CFile::GetSize64() {
    try {
        if (ExistsDir()) {
            vector<CFileName> vct = CFile::FindFile(GetName(), -1);
            uint64_t nSum = 0;
            for (size_t i = 0, n = vct.size(); i < n; i++) {
                nSum += CFindFile(vct[i].Get()).GetSize64();
            }
            return nSum;
        } else {
            return CFindFile(GetName().RemoveDirSeparator()).GetSize64();
        }
    } catch (std::exception&) {
        Themis_DebugWarning(L"CFile::GetSize() : " + GetName().Get() + L"݂Ȃ");
        return 0;
    }
}

FILETIME CFile::GetLastUpdate() {
    return CFindFile(GetName().RemoveDirSeparator()).GetLastUpdate();
}

/// t@C̃Rs[
void CFile::InnerCopy(const CFileName& oDstName, bool bOverwrite) {
    if (oDstName.IsDirPath()) {
        if (GetName().IsDirPath()) { // tH_tH_
            {vector<CFileName> vct = FindDir(GetName(), 0);
            for (size_t i = 0, n = vct.size(); i < n; i++) {
                CFile(vct[i]).InnerCopy(CFileName(oDstName.Get() + vct[i].GetDirName() + L"\\"), bOverwrite);
            }}
            {vector<CFileName> vct = FindFile(GetName(), 0);
            for (size_t i = 0, n = vct.size(); i < n; i++) {
                CFile(vct[i]).InnerCopy(oDstName, bOverwrite);
            }}
        } else { // t@CtH_
            InnerCopy(CFileName(oDstName.Get() + GetName().GetPureName()), bOverwrite);
        }
    } else { // t@Ct@C
        CFile(CFileName(oDstName.GetDir())).CreateDir();
        BOOL result = NativeString::IsUnicode() ?
            ::CopyFileW(GetName().Get().ToCharPtr(),
                oDstName.Get().ToCharPtr(), bOverwrite ? FALSE : TRUE) :
            ::CopyFileA(Encoding::GetDefault().GetBytes(GetName().Get()).c_str(),
                Encoding::GetDefault().GetBytes(oDstName.Get()).c_str(), bOverwrite ? FALSE : TRUE);
        if (result == FALSE) {
            Themis_DebugApiError();
            throw IOException(L"t@C̃Rs[Ɏs܂ (" + GetName().Get() + L"  " + oDstName.Get() + L")");
        }
    }
}

/// t@C̕ύX
void CFile::Move(const CFileName& oDstName) {
    if (oDstName.IsDirPath()) {
        if (GetName().IsDirPath()) { // tH_tH_
            {vector<CFileName> vct = FindDir(GetName(), 0);
            for (size_t i = 0, n = vct.size(); i < n; i++) {
                CFile(vct[i]).Move(CFileName(oDstName.Get() + vct[i].GetDirName() + L"\\"));
            }}
            {vector<CFileName> vct = FindFile(GetName(), 0);
            for (size_t i = 0, n = vct.size(); i < n; i++) {
                CFile(vct[i]).Move(oDstName);
            }}
            // Ȃ폜
            BOOL result = NativeString::IsUnicode() ?
                ::RemoveDirectoryW(GetName().RemoveDirSeparator().ToCharPtr()) :
                ::RemoveDirectoryA(Encoding::GetDefault().GetBytes(GetName().RemoveDirSeparator()).c_str());
            if (result == FALSE) {
                Themis_DebugApiError();
                throw IOException(GetName().Get() + L" ̍폜Ɏs܂");
            }
        } else { // t@CtH_
            Move(CFileName(oDstName.Get() + GetName().GetPureName()));
        }
    } else { // t@Ct@C
        CFile(CFileName(oDstName.GetDir())).CreateDir();
        BOOL result = NativeString::IsUnicode() ?
            ::MoveFileExW(GetName().RemoveDirSeparator().ToCharPtr(),
                oDstName.RemoveDirSeparator().ToCharPtr(),
                MOVEFILE_COPY_ALLOWED) :
            ::MoveFileExA(Encoding::GetDefault().GetBytes(GetName().RemoveDirSeparator()).c_str(),
                Encoding::GetDefault().GetBytes(oDstName.RemoveDirSeparator()).c_str(),
                MOVEFILE_COPY_ALLOWED);
        if (result == FALSE) {
            Themis_DebugApiError();
            throw IOException(L"t@C̈ړɎs܂ (" +
                GetName().Get() + L"  " + oDstName.Get() + L")");
        }
    }
}

void CFile::Delete(bool bDeleteForce) {
    if (ExistsDir()) {
        // ċAIɍ폜
        vector<CFileName> vct = Find(GetName(), 0);
        for (size_t i = 0; i < vct.size(); i++) {
            CFile(vct[i]).Delete(bDeleteForce);
        }
        // ̕ύX
        string oDirName = GetName().RemoveDirSeparator();
        //ǂ݂ƂpƂ̉
        if (bDeleteForce) {
            SetFileAttributes(oDirName, FILE_ATTRIBUTE_NORMAL);
        }
        // 폜
        BOOL result = NativeString::IsUnicode() ?
            ::RemoveDirectoryW(oDirName.ToCharPtr()) :
            ::RemoveDirectoryA(Encoding::GetDefault().GetBytes(oDirName).c_str());
        if (result == FALSE) {
            throw IOException(GetName().Get() + L"̍폜Ɏs܂");
        }
    } else {
        string strFullName = GetName().Get();
        //ǂ݂ƂpƂ̉
        if (bDeleteForce) {
            SetFileAttributes(strFullName, FILE_ATTRIBUTE_NORMAL);
        }
        BOOL result = NativeString::IsUnicode() ?
            ::DeleteFileW(strFullName.ToCharPtr()) :
            ::DeleteFileA(Encoding::GetDefault().GetBytes(strFullName).c_str());
        if (result == FALSE) {
            Themis_DebugApiError();
            throw IOException(GetName().Get() + L"̍폜Ɏs܂");
        }
    }
}

void CFile::DeleteEmpty() {
    Themis_DebugAssert(GetName().IsDirPath());
    if (ExistsDir()) {
        // ċAIɋfBNg폜
        vector<CFileName> vct = FindDir(GetName(), 0);
        for (size_t i = 0; i < vct.size(); i++) {
            CFile(vct[i]).DeleteEmpty();
        }
        // 󂶂ȂΎs邯ǖ
        NativeString::IsUnicode() ?
            ::RemoveDirectoryW(GetName().RemoveDirSeparator().ToCharPtr()) :
            ::RemoveDirectoryA(Encoding::GetDefault().GetBytes(GetName().RemoveDirSeparator()).c_str());
    }
}

void CFile::CreateDir() {
    if (!GetName().IsDirPath()) {
        CFile(CFileName(GetName().GetDir())).CreateDir();
    } else if (!ExistsDir()) {
        string strParentDir = GetName().GetParentDir();
        if (!string::IsNullOrEmpty(strParentDir)) {
            CFile(CFileName(strParentDir)).CreateDir(); //ċAătH_
        }
        BOOL result = NativeString::IsUnicode() ?
            ::CreateDirectoryW(GetName().Get().ToCharPtr(), NULL) :
            ::CreateDirectoryA(Encoding::GetDefault().GetBytes(GetName().Get()).c_str(), NULL);
        if (result == FALSE) {
            Themis_DebugApiError();
            throw IOException(GetName().Get() + L"̍쐬Ɏs܂");
        }
    }
}

bool CFile::IsReadOnly() {
    DWORD dwAttributes = GetFileAttributesSafe(GetName().Get());
    return (dwAttributes != 0xFFFFFFFF && (dwAttributes & FILE_ATTRIBUTE_READONLY) != 0);
}

void CFile::SetReadOnly(bool bOnly) {
    DWORD dwAttributes = GetFileAttributesSafe(GetName().Get());
    Themis_DebugAssert(dwAttributes != 0xFFFFFFFF);
    if (bOnly) {
        dwAttributes |= FILE_ATTRIBUTE_READONLY;
    } else {
        dwAttributes &= ~FILE_ATTRIBUTE_READONLY;
    }
    SetFileAttributes(GetName().Get(), dwAttributes);
}

void CFile::Load(vector<uint8_t>& rMemory) {
    CStdFile oFile(GetName(), L"rb");
    size_t n = oFile.GetSize();
    if (n <= 0) {
        rMemory.clear();
    } else {
        rMemory.clear();
        rMemory.resize(n);
        oFile.Read(&rMemory[0], n);
    }
}

void CFile::Load(StringBuilder& str) {
    CStdFile file(GetName(), L"rb");
    size_t n = file.GetSize();
    str.Remove(0, str.GetLength());
    if (0 < n) {
        BOOST_STATIC_ASSERT(sizeof(wchar_t) == 2);
        if (2 < n && n % 2 == 0) { // BOMtŃTCY̏ꍇ
            uint16_t bom;
            file.Read(&bom, sizeof(bom));
            if (bom == 0xfeff) {
                // Unicode (lCeBuȃGfBA)
                str.SetLength(static_cast<int>(n / 2 - 1));
                file.Read(&str[0], n - 2);
                return;
            } else if (bom == 0xfffe) {
                // Unicode (lCeBuȂGfBA)
                str.SetLength(static_cast<int>(n / 2 - 1));
                file.Read(&str[0], n - 2);
                // oCgI[_[̃Xbv
                for (int i = 0, m = str.GetLength(); i < m; i++) {
                    uint16_t n = static_cast<uint16_t>(str[i]);
                    str[i] = static_cast<wchar_t>(
                        ((n & 0x00ff) << 8) | 
                        ((n & 0xff00) >> 8));
                }
                return;
            } else {
                // BOMUnicodeȂȂ̂ACPƂĂ݂
                file.Seek(0, SEEK_SET);
            }
        }
        std::string data(n, '\0');
        file.Read(&data[0], n);
        str.Append(Encoding::GetDefault().GetString(data));
    }
}

void CFile::Save(const void* p, size_t n) {
    CreateDir();
    CStdFile(GetName(), L"w+b").Write(p, n);
}

void CFile::Save(const vector<uint8_t>& vct) {
    Save(vct.empty() ? NULL : &vct[0], vct.size());
}

void CFile::Save(const string& str) {
    CreateDir();
    CStdFile file(GetName(), L"w+b");
    uint16_t bom = 0xfeff;
    file.Write(&bom, sizeof(bom));
    file.Write(str.ToCharPtr(), str.GetLength() * 2);
    BOOST_STATIC_ASSERT(sizeof(wchar_t) == 2);
}

vector<CFileName> CFile::FindFile(const CFileName& oDirName, int nReflexive) {
    Themis_DebugAssert(oDirName.IsDirPath());
    string strFullDir = oDirName.GetDir();
    CFindFile oFindFile(strFullDir + L"*");
    vector<CFileName> o;
    do {
        if ((oFindFile.GetAttributes() & FILE_ATTRIBUTE_DIRECTORY) != 0) {
            //fBNgȏꍇ
            if (0 != nReflexive) {   //ċAꍇ
                if (!oFindFile.IsDot()) {
                    try {
                        vector<CFileName> v = FindFile(CFileName(oFindFile.GetFullName() + L"\\"), nReflexive - 1);
                        o.insert(o.end(), v.begin(), v.end());
                    } catch (std::exception&) {
                        Themis_DebugWarning(L"t@CɃG[");
                    }
                }
            }
        } else {
            //t@Cȏꍇ
            o.push_back(CFileName(oFindFile.GetFullName()));
        }
    } while (oFindFile.ToNext());
    return o;
}

vector<CFileName> CFile::FindDir(const CFileName& oDirName, int nReflexive) {
    Themis_DebugAssert(oDirName.IsDirPath());
    string strFullDir = oDirName.GetDir();
    CFindFile oFindFile(strFullDir + L"*");
    vector<CFileName> o;
    do {
        if ((oFindFile.GetAttributes() & FILE_ATTRIBUTE_DIRECTORY) != 0) {
            //fBNgȏꍇ
            if (!oFindFile.IsDot()) {
                o.push_back(CFileName(oFindFile.GetFullName() + L"\\"));
                if (0 != nReflexive) {   //ċAꍇ
                    try {
                        vector<CFileName> v = FindDir(o[o.size() - 1], nReflexive-1);
                        o.insert(o.end(), v.begin(), v.end());
                    } catch (std::exception&) {
                        Themis_DebugWarning(L"t@CɃG[");
                    }
                }
            }
        } else {
            //t@Cȏꍇ
        }
    } while (oFindFile.ToNext());
    return o;
}

vector<CFileName> CFile::Find(const CFileName& oDirName, int nReflexive) {
    Themis_DebugAssert(oDirName.IsDirPath());
    string strFullDir = oDirName.GetDir();
    CFindFile oFindFile(strFullDir + L"*");
    vector<CFileName> o;
    do {
        if ((oFindFile.GetAttributes() & FILE_ATTRIBUTE_DIRECTORY) != 0) {
            //fBNgȏꍇ
            if (!oFindFile.IsDot()) {
                o.push_back(CFileName(oFindFile.GetFullName() + L"\\"));
                if (0 != nReflexive) {   //ċAꍇ
                    try {
                        vector<CFileName> v = Find(o[o.size() - 1], nReflexive - 1);
                        o.insert(o.end(), v.begin(), v.end());
                    } catch (std::exception&) {
                        Themis_DebugWarning(L"t@CɃG[");
                    }
                }
            }
        } else {
            //t@Cȏꍇ
            o.push_back(CFileName(oFindFile.GetFullName()));
        }
    } while (oFindFile.ToNext());
    return o;
}

/// oDirName݂ȂȂ̊̕Kw֍sA݂tH_ԂB[g瑶݂Ȃꍇ͋󕶎B
CFileName CFile::GetExistDirName(const CFileName& oDirName) {
    Themis_DebugAssert(oDirName.IsEmpty() || !oDirName.IsRelativePath());
    Themis_DebugAssert(oDirName.IsDirPath());
    CFileName dir = oDirName;
    while (!string::IsNullOrEmpty(dir.Get())) {
        if (CFile(dir).ExistsDir()) {
            break;
        } else {
            dir = CFileName(dir.GetParentDir());
        }
    }
    return dir;
}

/// I[tH_؂蕶ł͂ȂA݂tH_̖OȏꍇA؂蕶ǉ̂Ԃ
CFileName CFile::SetDirSeparator(const CFileName& oFileName) {
    if (!oFileName.IsDirPath()) {
        DWORD dwAttributes = GetFileAttributesSafe(oFileName.Get());
        if (dwAttributes != 0xFFFFFFFF && (dwAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0) {
            return CFileName(oFileName.Get() + L"\\");
        }
    }
    return oFileName;
}

CFile::SFindInfo::SFindInfo(const CFindFile& rFindFile) {
    oFileName = rFindFile.GetFullFileName();
    nFileSize = rFindFile.GetSize64();
    nAttribute = rFindFile.GetAttributes();
    ftLastUpdate = rFindFile.GetLastUpdate();
    bDir = rFindFile.IsDir();
}

}
