#include "../../Windows/NativeString.h"
namespace Themis {

CDC::CDC(HDC hdc): m_hdc(hdc), m_szPlane(CSize::GetLarge()) {
/*
    HBITMAP hBmp = static_cast<HBITMAP>(::GetCurrentObject(GetHandle(), OBJ_BITMAP));
    if (hBmp != NULL) {
        BITMAP oBmp = {0};
        ::GetObject(hBmp, sizeof(oBmp), &oBmp);
        if (1 < oBmp.bmWidth && 1 < oBmp.bmHeight) {
            //擾oȂݒ肵Ƃ
            m_szPlane.cx = abs(oBmp.bmWidth);
            m_szPlane.cy = abs(oBmp.bmHeight);
        } else {
            Themis_DebugWarning(L"HDC̃TCY擾Ɏs");
        }
    } else {
        Themis_DebugWarning(L"HDC̃TCY擾Ɏs");
    }
*/
}

CDC::CDC(HDC hdc, const CSize& szPlane): m_hdc(hdc), m_szPlane(szPlane) {
}

CDC::CDC(CPlane::CGdiLock& r): m_hdc(r.Get()), m_szPlane(CSize::GetLarge()) {
}

CDC::~CDC() {
}

int CDC::GetBpp() const {
    return ::GetDeviceCaps(GetHandle(), BITSPIXEL);
}

void CDC::Fill(COLORREF rgb) {
    Fill(rgb, CRect(GetSize()));
}

void CDC::Fill(COLORREF rgb, const CRect& rc) {
    CRect rc2 = rc.GetIntersect(CRect(GetSize()));
    LOGBRUSH lbr = {0};
    lbr.lbStyle = BS_SOLID;
    lbr.lbColor = rgb;
    HBRUSH hbr = ::CreateBrushIndirect(&lbr);
    ::FillRect(GetHandle(), &rc2, hbr);
    ::DeleteObject(hbr);
}

void CDC::FillBlack() {
    FillBlack(CRect(GetSize()));
}

void CDC::FillWhite() {
    FillWhite(CRect(GetSize()));
}

void CDC::FillBlack(const CRect& rc) {
    ::PatBlt(GetHandle(), rc.left, rc.top, rc.GetWidth(), rc.GetHeight(), BLACKNESS);
}

void CDC::FillWhite(const CRect& rc) {
    ::PatBlt(GetHandle(), rc.left, rc.top, rc.GetWidth(), rc.GetHeight(), WHITENESS);
}

void CDC::ReverseFill(const CRect& rc) {
    ::PatBlt(GetHandle(), rc.left, rc.top, rc.GetWidth(), rc.GetHeight(), DSTINVERT);
}

void CDC::Blt(const CDC& rSrc, const CClipper& rClip) {
    Blt(rSrc, CPoint(), CRect(GetSize()), rClip);
}

void CDC::Blt(const CDC& rSrc, CPoint ptDst, CRect rcSrc, const CClipper& rClip) {
    if (rClip.Clipping(ptDst, rcSrc)) {
        BOOL bRet = ::BitBlt(GetHandle(), ptDst.x, ptDst.y, rcSrc.GetWidth(), rcSrc.GetHeight(), rSrc.GetHandle(), rcSrc.left, rcSrc.top, SRCCOPY);
        Themis_DebugAssertApi(bRet != FALSE);
    }
}

void CDC::CaptureBlt(const CDC& rSrc, CPoint ptDst, CRect rcSrc, const CClipper& rClip) {
    Themis_DebugLog(L"WINVER >= 0x0500 Kv CAPTUREBLT gp܂");
    const int CAPTUREBLT = 0x40000000;
    if (rClip.Clipping(ptDst, rcSrc)) {
        BOOL bRet = ::BitBlt(GetHandle(), ptDst.x, ptDst.y, rcSrc.GetWidth(), rcSrc.GetHeight(), rSrc.GetHandle(), rcSrc.left, rcSrc.top, SRCCOPY | CAPTUREBLT);
        Themis_DebugAssertApi(bRet != FALSE);
    }
}

void CDC::Line(const CPoint& pt1, const CPoint& pt2, COLORREF rgb, int nFat) {
    LOGPEN lp;
    lp.lopnStyle = PS_SOLID;
    lp.lopnWidth.x = nFat;
    lp.lopnWidth.y = 0;
    lp.lopnColor = rgb;
    HPEN hPen = ::CreatePenIndirect(&lp);
    HPEN hPenOld = (HPEN)::SelectObject(GetHandle(), hPen);
    ::MoveToEx(GetHandle(), pt1.x, pt1.y, NULL);
    ::LineTo(GetHandle(), pt2.x, pt2.y);
    ::SelectObject(GetHandle(), hPenOld);
    ::DeleteObject(hPen);
}

void CDC::StretchBltIndirect(const CDC& rSrc, const CRect& rcDst, const CRect& rcSrc, const CClipper& rClip) {
    CRect rcDst2 = rcDst;
    CRect rcSrc2 = rcSrc;
    if (rClip.Clipping(rcDst2, rcSrc2)) {
        BOOL bRet = ::StretchBlt(GetHandle(), rcDst2.left, rcDst2.top, rcDst2.GetWidth(), rcDst2.GetHeight(),
            rSrc.GetHandle(), rcSrc2.left, rcSrc2.top, rcSrc2.GetWidth(), rcSrc2.GetHeight(), SRCCOPY);
        Themis_DebugAssertApi(bRet != FALSE);
    }
}

void CDC::StretchBltNearest(const CDC& rSrc, const CRect& rcDst, const CRect& rcSrc, const CClipper& rClip) {
    CRect rcDst2 = rcDst;
    CRect rcSrc2 = rcSrc;
    if (rClip.Clipping(rcDst2, rcSrc2)) {
        int n = ::SetStretchBltMode(GetHandle(), COLORONCOLOR);
        Themis_DebugAssertApi(n != 0);
        BOOL bRet = ::StretchBlt(GetHandle(), rcDst2.left, rcDst2.top, rcDst2.GetWidth(), rcDst2.GetHeight(),
            rSrc.GetHandle(), rcSrc2.left, rcSrc2.top, rcSrc2.GetWidth(), rcSrc2.GetHeight(), SRCCOPY);
        Themis_DebugAssertApi(bRet != FALSE);
        ::SetStretchBltMode(GetHandle(), n);
    }
}

void CDC::StretchBlt(const CDC& rSrc, const CRect& rcDst, const CRect& rcSrc, const CClipper& rClip) {
    if (NativeString::IsUnicode()) {
        CRect rcDst2 = rcDst;
        CRect rcSrc2 = rcSrc;
        if (rClip.Clipping(rcDst2, rcSrc2)) {
            int n = ::SetStretchBltMode(GetHandle(), HALFTONE);
            Themis_DebugAssertApi(n != 0);
            ::SetBrushOrgEx(GetHandle(), 0, 0, NULL); //Ȃ񂩂悭񂯂MSDNɏĂ̂ŁB
            BOOL bRet = ::StretchBlt(GetHandle(), rcDst2.left, rcDst2.top, rcDst2.GetWidth(), rcDst2.GetHeight(),
                rSrc.GetHandle(), rcSrc2.left, rcSrc2.top, rcSrc2.GetWidth(), rcSrc2.GetHeight(), SRCCOPY);
            Themis_DebugAssertApi(bRet != FALSE);
            ::SetStretchBltMode(GetHandle(), n);
        }
    } else {
        StretchBltNearest(rSrc, rcDst, rcSrc, rClip);
    }
}

/// ƂpAStretchBltg邩ǂB
bool CDC::IsStretchBltSupport() const {
    int n = ::GetDeviceCaps(GetHandle(), RASTERCAPS);
    return (n & RC_BITMAP64) && (n & RC_STRETCHBLT);
}

}
