#include "stdafx.h"
#include <olectl.h>
#pragma comment(lib, "olepro32.lib")
#include "../File.h"
#include "../Graphics.h"
#include "OleLoader.h"
namespace Themis {

COleLoader::COleLoader(): m_pPicture(NULL) {
}

COleLoader::~COleLoader() {
    if (m_pPicture != NULL) {
        static_cast<IPicture*>(m_pPicture)->Release();
    }
}

bool COleLoader::Load(const std::tr1::shared_ptr<CFileData>& fileData, bool bRenderOnlyAtOnce) {
    const vector<uint8_t>& fileMemory = fileData->GetMemory();
    class CMemory {
        HGLOBAL m_hGlobal;
        void* m_pMemory;
    public:
        CMemory(size_t nSize) {
            m_hGlobal = ::GlobalAlloc(GMEM_MOVEABLE, nSize);
            m_pMemory = ::GlobalLock(m_hGlobal);
        }
        ~CMemory() { ::GlobalFree(m_hGlobal); }
        void UnLock() { ::GlobalUnlock(m_hGlobal); }
        HGLOBAL GetHandle() { return m_hGlobal; }
        void* GetPtr() { return m_pMemory; }
    };
    CMemory oMemory(fileMemory.size());
    memcpy(oMemory.GetPtr(), &fileMemory[0], fileMemory.size());
    oMemory.UnLock();

    ::IStream* pStream;
    HRESULT ret = ::CreateStreamOnHGlobal(oMemory.GetHandle(), FALSE, &pStream);
    if (S_OK != ret) return false;

    ret = ::OleLoadPicture(pStream, LONG(fileMemory.size()), TRUE, IID_IPicture, &m_pPicture);
    pStream->Release();
    if (ret != S_OK) return false;

    // TODO: ߐF̗LƂׂ̂BB

    return true;
}

CSize COleLoader::GetSize() const {
    OLE_XSIZE_HIMETRIC cx;
    OLE_XSIZE_HIMETRIC cy;
    static_cast<IPicture*>(m_pPicture)->get_Width(&cx);
    static_cast<IPicture*>(m_pPicture)->get_Height(&cy);
    HDC hdc = ::GetDC(NULL);
    int x = ::MulDiv(cx, ::GetDeviceCaps(hdc, LOGPIXELSX), 2540);
    int y = ::MulDiv(cy, ::GetDeviceCaps(hdc, LOGPIXELSY), 2540);
    ::ReleaseDC(NULL, hdc);
    return CSize(x, y);
}

void COleLoader::Render(CPlane& dst, const CPoint& pt, bool bRenderOnlyAtOnce) {
    OLE_XSIZE_HIMETRIC cx;
    OLE_XSIZE_HIMETRIC cy;
    static_cast<IPicture*>(m_pPicture)->get_Width(&cx);
    static_cast<IPicture*>(m_pPicture)->get_Height(&cy);
    HDC hdc = ::GetDC(NULL);
    int w = ::MulDiv(cx, ::GetDeviceCaps(hdc, LOGPIXELSX), 2540);
    int h = ::MulDiv(cy, ::GetDeviceCaps(hdc, LOGPIXELSY), 2540);
    ::ReleaseDC(NULL, hdc);

    HRESULT ret = static_cast<IPicture*>(m_pPicture)->Render(
        CPlane::CGdiLock(dst), pt.x, pt.y, w, h, 0, cy, cx, -cy, NULL);
    if (ret != S_OK) { throw ImageLoadException(); }
}

string COleLoader::GetTypeName() const {
    short type;
    if (m_pPicture == NULL || S_OK != static_cast<IPicture*>(m_pPicture)->get_Type(&type)) {
        return L"OLE Image";
    } else {
        switch (type) {
        case PICTYPE_BITMAP:        return L"OLE rbg}bv";
        case PICTYPE_METAFILE:      return L"OLE ^t@C";
        case PICTYPE_ICON:          return L"OLE ACR";
        case PICTYPE_ENHMETAFILE:   return L"OLE g^t@C";
        default:                    return L"OLE Image";
        }
    }
}

}
