#include "stdafx.h"
#include "ThreadedImageLoader.h"
#include "../Thread/Thread.h"
#include "../Thread/ThreadPool.h"
namespace Themis {

ThreadedImageLoader::ThreadedImageLoader(const std::tr1::shared_ptr<IOwner>& o) : owner(o), cachePos(0) {
    // XbhJn
    threadValid = true;
    thread = ThreadPool::CreateWorkerThread();
    thread->SetWork(std::tr1::bind(&ThreadedImageLoader::OnThread, this));
}

ThreadedImageLoader::~ThreadedImageLoader() {
    {
        boost::mutex::scoped_lock lock(mutex);
        threadValid = false;
        condition.notify_all();
    }
    thread->Join();
}

/// ő̐ݒ(ǂݍ݂̓LZ)
void ThreadedImageLoader::SetCacheCount(int n) {
    boost::mutex::scoped_lock lock(mutex);
    if (static_cast<size_t>(n) != cache.size()) {
        if (n < 3) n = 0;
        cache.clear();
        cache.resize(n);
        cachePos = 0;
        m_bCancel = true;
        loadQueue.clear();
    }
}

/// ǂ݂SăLZALbVNA
void ThreadedImageLoader::Clear() {
    boost::mutex::scoped_lock lock(mutex);
    m_bCancel = true;
    loadQueue.clear();
    for (size_t i = 0, n = cache.size(); i < n; i++) cache[i].reset();
    // ǂݒÔߏȊ҂Ƃ
    if (loadingEntry != NULL) {
        loadingCondition.wait(lock);
    }
}

/// ǂݍ݃X^bNɐς
void ThreadedImageLoader::PreLoad(const std::tr1::shared_ptr<IFileID>& fileID) {
    boost::mutex::scoped_lock lock(mutex);
    // LbV[hB
    if (cache.empty()) {
        return;
    }
    // ǂݍݐ^ŒȂ疳
    if (loadingEntry != NULL && loadingEntry->fileID->Equals(*fileID)) {
        return;
    }
    // LbVɂȂ疳
    for (size_t i = 0, n = cache.size(); i < n; i++) {
        if (cache[i].get() != NULL && cache[i]->fileID->Equals(*fileID)) {
            return;
        }
    }
    // ǂݑ҂ł
    for (std::list<std::tr1::shared_ptr<SEntry> >::iterator it = loadQueue.begin(), itEnd = loadQueue.end(); it != itEnd; ++it) {
        if ((*it)->fileID->Equals(*fileID)) {
            return;
        }
    }
    // X^bNɐς
    if (0 < cache.size()) {
        std::tr1::shared_ptr<SEntry> s(new SEntry());
        s->fileID = fileID;
        loadQueue.push_back(s);
        condition.notify_all();
    }
}

/// ǂݍ
std::tr1::shared_ptr<CImageLoader> ThreadedImageLoader::Load(const std::tr1::shared_ptr<IFileID>& fileID) {
    {
        boost::mutex::scoped_lock lock(mutex);
        // ǂݍݐ^ŒȂÎ҂B
        if (loadingEntry != NULL && loadingEntry->fileID->Equals(*fileID)) {
            Themis_DebugLog(L"ThreadedImageLoader.Load: LbVǂݍݑ҂");
            loadingCondition.wait(lock);
        }
        // LbV[hB
        if (cache.empty()) {
            lock.unlock();
            // ̏œǂݍ
            std::tr1::shared_ptr<SEntry> s(new SEntry());
            s->fileID = fileID;
            InnerLoad(s, true);
            return s->loader;
        }
        // LbVɂȂ炻Ԃ
        for (size_t i = 0, n = cache.size(); i < n; i++) {
            if (cache[i].get() != NULL && cache[i]->fileID->Equals(*fileID)) {
                return cache[i]->loader; // I
            }
        }
        Themis_DebugLog(L"ThreadedImageLoader.Load: LbV~X");
        // ǂݑ҂ȂLZ
        for (std::list<std::tr1::shared_ptr<SEntry> >::iterator it = loadQueue.begin(), itEnd = loadQueue.end(); it != itEnd; ) {
            if ((*it)->fileID->Equals(*fileID)) {
                loadQueue.erase(it++);
            } else {
                ++it;
            }
        }
    }
    // ̏œǂݍ
    std::tr1::shared_ptr<SEntry> s(new SEntry());
    s->fileID = fileID;
    InnerLoad(s, false);
    boost::mutex::scoped_lock lock(mutex);
    size_t n = InnerAddCache(s);
    return cache[n]->loader;
}

void ThreadedImageLoader::OnThread() {
    Thread::SetName(L"ThreadedImageLoaderThread");
    while (true) {
        {
            boost::mutex::scoped_lock lock(mutex);
            if (!threadValid) return;
            while (loadQueue.empty()) {
                // Ȃҋ@
                condition.wait(lock);
                if (!threadValid) return;
            }
            // ǂݍݏ
            loadingEntry = loadQueue.front();
            loadQueue.pop_front();
            m_bCancel = false;
        }
        // ǂݍ
        InnerLoad(loadingEntry, false);
        // ǂݍ݌̃R[obN
        {
            boost::mutex::scoped_lock lock(mutex);
            if (!threadValid) return;
            if (!m_bCancel) {
                InnerAddCache(loadingEntry);
            }
            loadingEntry.reset();
            loadingCondition.notify_all();
        }
    }
}

}
