#pragma once
#include "Image.h"
#include "../Factory.h"
#include "../String/Encoding.h"
namespace Themis {

class WorkerThreadController;

/**
 * ʃXbhŉ摜̓ǂݍݏsIuWFNgB
 */
class ThreadedImageLoader : private boost::noncopyable {
public:
    /// t@Cf[^IuWFNg
    struct IFileID {
        virtual ~IFileID() {}
        virtual bool Equals(const IFileID& r) const = 0;
    };
    /// FXR[obNpC^[tF[X
    class IOwner {
    public:
        virtual ~IOwner() {}
        virtual std::tr1::shared_ptr<CImageLoader> CreateImageLoader() = 0;
        virtual std::tr1::shared_ptr<CFileData> LoadFile(ThreadedImageLoader::IFileID& fileID) = 0;
    };

    ThreadedImageLoader(const std::tr1::shared_ptr<IOwner>& owner);
    ~ThreadedImageLoader();

    /// ő̐ݒ(ǂݍ݂̓LZ)BRȂLbV[hB
    void SetCacheCount(int n);
    /// LbV̎擾
    int GetCacheCount() const { return static_cast<int>(cache.size()); }
    /// LbV[hȂtrue
    bool IsCacheless() const { return cache.empty(); }

    /// ǂ݂SăLZALbVNA
    void Clear();

    /// ǂ݃X^bNɐς
    void PreLoad(const std::tr1::shared_ptr<IFileID>& fileID);

    /// ǂݍ݁BsnullԂĂ
    std::tr1::shared_ptr<CImageLoader> Load(const std::tr1::shared_ptr<IFileID>& fileID);

private:
    void OnThread();

    struct SEntry {
        std::tr1::shared_ptr<IFileID> fileID;
        std::tr1::shared_ptr<CImageLoader> loader;
    };

    void InnerLoad(const std::tr1::shared_ptr<SEntry>& s, bool renderOnlyAtOnce) {
        s->loader = owner->CreateImageLoader();
        Themis_DebugAssert(renderOnlyAtOnce || !s->loader->IsRenderOnlyAtOnce());
        try {
            s->loader->SetRenderOnlyAtOnce(renderOnlyAtOnce);
            s->loader->Load(owner->LoadFile(*s->fileID));
        } catch (std::exception& e) {
            Themis_DebugLog(L"ThreadedImageLoader.InnerLoad: " + Encoding::GetDefault().GetString(e.what())); e;
            s->loader.reset();
        }
    }
    size_t InnerAddCache(const std::tr1::shared_ptr<SEntry>& s) {
        size_t n = cachePos;
        cache[cachePos] = s;
        cachePos = (cachePos + 1) % cache.size();
        return n;
    }

private:
    std::tr1::shared_ptr<IOwner> owner;
    vector<std::tr1::shared_ptr<SEntry> > cache;
    size_t cachePos;

    volatile bool threadValid;
    std::tr1::shared_ptr<WorkerThreadController> thread;
    boost::mutex mutex;
    boost::condition condition;

    boost::condition loadingCondition;
    std::tr1::shared_ptr<SEntry> loadingEntry;

    std::list<std::tr1::shared_ptr<SEntry> > loadQueue;
    volatile bool m_bCancel;
};


}
