#include "stdafx.h"
#include "Socket.h"
#include "StringScanner.h"
#include "String/Encoding.h"
#include "Utility/StringUtility.h"
#include <winsock.h>
#pragma comment(lib, "ws2_32.lib")
namespace Themis {

struct STcpSocketData {
    SOCKET              oSocket;
};

namespace {
    class CSocketInitializer {
        WSADATA     m_wd;
    public:
        CSocketInitializer() {
            int nErrorStatus = ::WSAStartup(MAKEWORD(1, 1), &m_wd);
            if (nErrorStatus != 0) {
                throw RuntimeException(L"WinSock̏Ɏs܂");
            }
        }
        ~CSocketInitializer() {
            ::WSACleanup();
        }
    };
    void InitializeSocket() {
        static CSocketInitializer s_oInitializer;
    }
}

CTcpSocket::CTcpSocket() {
    InitializeSocket();
    // fXgN^̌Ă΂ꏇIɁA肠Ƃ
    // CTcpSocket̂鎖ۏoȂȂǁc
    data->oSocket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (data->oSocket == INVALID_SOCKET) {
        throw RuntimeException(L"\Pbg̍쐬Ɏs܂");
    }
}

CTcpSocket::~CTcpSocket() {
    if (data->oSocket != INVALID_SOCKET) {
        int r1 = shutdown(data->oSocket, 2);
        int r2 = closesocket(data->oSocket);
    /*
        Themis_DebugAssert(r1 != SOCKET_ERROR);
        Themis_DebugAssert(r2 != SOCKET_ERROR);
    //*/
    }
}

void CTcpSocket::Bind(unsigned short nPort) {
    sockaddr_in oServer = {0};
    oServer.sin_family = AF_INET;
    oServer.sin_port = htons(nPort);
    if (bind(data->oSocket, reinterpret_cast<sockaddr*>(&oServer), sizeof(oServer)) == SOCKET_ERROR) {
        throw RuntimeException(ItoA(nPort) + L"ԃ|[gւ̃\Pbg̊֘AtɎs܂");
    }
}

void CTcpSocket::Listen(int backlog) {
    listen(data->oSocket, backlog);
}

namespace {
    class CTcpConnection : public ITcpConnection {
        SOCKET m_oSocket;
    public:
        CTcpConnection(SOCKET s) : m_oSocket(s) {
        }
        ~CTcpConnection() {
            if (m_oSocket != INVALID_SOCKET) {
                int r1 = shutdown(m_oSocket, 2);
                int r2 = closesocket(m_oSocket);
                Themis_DebugAssert(r1 != SOCKET_ERROR);
                Themis_DebugAssert(r2 != SOCKET_ERROR);
            }
        }
        void Send(const void* p, size_t n) {
            Themis_DebugAssert(m_oSocket != INVALID_SOCKET);
            if (send(m_oSocket, reinterpret_cast<const char*>(p), static_cast<int>(n), 0) == SOCKET_ERROR) {
                throw RuntimeException(L"NCAgւ̑MɎs܂");
            }
        }
        int Receive(void* p, size_t n) {
            Themis_DebugAssert(m_oSocket != INVALID_SOCKET);
            int ret = recv(m_oSocket, reinterpret_cast<char*>(p), static_cast<int>(n), 0);
            if (ret == SOCKET_ERROR) {
                throw RuntimeException(L"NCAg̎MɎs܂");
            }
            return ret;
        }
    };
}
auto_ptr<ITcpConnection> CTcpSocket::Accept() {
    SOCKADDR_IN oClient;
    int nClientAddr = sizeof(SOCKADDR_IN);
    SOCKET s = accept(data->oSocket, reinterpret_cast<SOCKADDR*>(&oClient), &nClientAddr);
    return auto_ptr<ITcpConnection>(new CTcpConnection(s));
}

void CTcpSocket::Close() {
    if (data->oSocket != INVALID_SOCKET) {
        SOCKET s = data->oSocket;
        data->oSocket = INVALID_SOCKET;
        shutdown(s, 2);
        closesocket(s);
    }
}

namespace {
    void GetServerAddr(unsigned long& rServerAddr, unsigned short& rPort, const string& strServer, unsigned short nDefaultPort) {
        // u:v؂Ń|[gԍꍇAT[o[ƕ
        string strServerName;
        rPort = nDefaultPort;
        List<string> serverParts = strServer.Split(Array(L':'), 2);
        if (serverParts.GetCount() == 2) {
            strServerName = serverParts[0];
            rPort = static_cast<unsigned short>(AtoI(serverParts[1]));
        } else {
            strServerName = strServer;
        }
        // IPAhX̉
        std::string mb = Encoding::GetDefault().GetBytes(strServerName);
        rServerAddr = inet_addr(const_cast<char*>(mb.c_str()));
        if (rServerAddr == -1) {
            // T[o[T[ȍ擾
            hostent* pHostEntry = gethostbyname(mb.c_str());
            if (pHostEntry == NULL) {
                throw RuntimeException(strServer + L" IP̎擾Ɏs܂");
            }
            rServerAddr = *reinterpret_cast<unsigned long*>((pHostEntry->h_addr_list)[0]);
        }
    }
}

void CTcpSocket::Connect(const string& strServer, unsigned short nDefaultPort) {
    unsigned long nServerAddr;
    unsigned short nPort;
    GetServerAddr(nServerAddr, nPort, strServer, nDefaultPort);
    // T[o[֐ڑ
    sockaddr_in oServer = {0};
    oServer.sin_family = AF_INET;
    oServer.sin_addr.s_addr = nServerAddr;
    oServer.sin_port = htons(nPort);
    if (connect(data->oSocket, reinterpret_cast<sockaddr*>(&oServer), sizeof(oServer)) == SOCKET_ERROR) {
        throw RuntimeException(strServer + L"ւ̐ڑɎs܂");
    }
}

void CTcpSocket::Send(const std::string& str) {
    //O\Ă݂
    Themis_DebugLog(L"--- CTcpSocket::Send() ---");
    Themis_DebugLog(Encoding::GetDefault().GetString(str));
    Themis_DebugLog(L"-----------------------");
    Send(str.c_str(), str.length());
}

void CTcpSocket::Send(const void* _p, size_t _n) {
    const char* p = static_cast<const char*>(_p);
    int n = static_cast<int>(_n);
    while (0 < n) {
        int nSent = send(data->oSocket, p, n, 0);
        if (nSent <= 0) {
            Themis_DebugAssert(nSent == SOCKET_ERROR);
            throw RuntimeException(L"T[o[ւ̑MɎs܂");
        }
        p += nSent;
        n -= nSent;
    }
    Themis_DebugAssert(n == 0);
}

int CTcpSocket::Receive(void* p, size_t n) {
    int nRet = recv(data->oSocket, reinterpret_cast<char*>(p), static_cast<int>(n), 0);
    if (nRet == SOCKET_ERROR) {
        throw RuntimeException(L"T[o[̎MɎs܂");
    }
    return nRet;
}

void CTcpSocket::ReceiveLine(string& str) {
    str = string::Empty;
    char sz[4096];
    while (true) {
        int n = Receive(sz, sizeof(sz) - 1); // L'\0'̕PȂƂƁBB
        if (n != 0) {
            sz[n] = L'\0';
            str += Encoding::GetDefault().GetString(sz);
            if (str.Contains(L"\n")) {
                break;
            }
        } else {
            break;
        }
    }
    //O\Ă݂
    Themis_DebugLog(L"--- CTcpSocket::ReceiveLine() ---");
    Themis_DebugLog(str);
    Themis_DebugLog(L"--------------------------");
}

//------------------------------------------------------------------//

struct SUdpHostData {
    sockaddr_in oHost;
};

struct SUdpSocketData {
    SOCKET oSocket;
};

CUdpHost::CUdpHost(const string& strServer, unsigned short nDefaultPort) {
    InitializeSocket();
    unsigned long nServerAddr;
    unsigned short nPort;
    GetServerAddr(nServerAddr, nPort, strServer, nDefaultPort);
    data->oHost.sin_family = AF_INET;
    data->oHost.sin_addr.s_addr = nServerAddr;
    data->oHost.sin_port = htons(nPort);
    memset(data->oHost.sin_zero, 0, sizeof(data->oHost.sin_zero));
}

CUdpHost::~CUdpHost() {
}

void* CUdpHost::GetAddr() const {
    return &data->oHost;
}

CUdpSocket::CUdpSocket() {
    InitializeSocket();
    data->oSocket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (data->oSocket == INVALID_SOCKET) {
        throw RuntimeException(L"\Pbg̍쐬Ɏs܂");
    }
}

CUdpSocket::~CUdpSocket() {
    if (data->oSocket != INVALID_SOCKET) {
        shutdown(data->oSocket, 2);
        closesocket(data->oSocket);
    }
}

void CUdpSocket::Bind() {
    sockaddr_in oClient = {0};
    oClient.sin_family = AF_INET;
    oClient.sin_addr.s_addr = INADDR_ANY;
    oClient.sin_port = htons(0);
    if (bind(data->oSocket, reinterpret_cast<sockaddr*>(&oClient), sizeof(oClient)) == SOCKET_ERROR) {
        throw RuntimeException(L"\PbgbindɎs܂");
    }
}

bool CUdpSocket::Select(int nTimeout) {
    timeval tv;
    tv.tv_sec = nTimeout;
    tv.tv_usec = 0;
    fd_set fs;
    FD_ZERO(&fs);
    FD_SET(data->oSocket, &fs);
    int nRet = select(FD_SETSIZE, &fs, 0, 0, &tv);
    if (nRet == SOCKET_ERROR) {
        throw RuntimeException(L"\PbgʐMɃG[܂");
    } else if (nRet == 0) {
        return false; // ^CAEg
    }
    if (!FD_ISSET(data->oSocket, &fs)) {
        throw RuntimeException(L"\PbgʐMɃG[܂");
    }
    return true;
}

void CUdpSocket::SendTo(const CUdpHost& rHost, const void* p, size_t n) {
    if (sendto(data->oSocket, reinterpret_cast<const char*>(p), static_cast<int>(n), 0, static_cast<sockaddr*>(rHost.GetAddr()), sizeof(sockaddr_in)) == SOCKET_ERROR) {
        throw RuntimeException(L"T[o[ւ̃NGXg̑MɎs܂");
    }
}

int CUdpSocket::ReceiveFrom(const CUdpHost& rHost, void* p, size_t n) {
    int nSize = (int)sizeof(sockaddr_in);
    int ret = recvfrom(data->oSocket, reinterpret_cast<char*>(p), static_cast<int>(n), 0, static_cast<sockaddr*>(rHost.GetAddr()), &nSize);
    if (ret == SOCKET_ERROR) {
        throw RuntimeException(L"T[o[̎MɎs܂");
    }
    return nSize; //return ret;
}

}
