#include "stdafx.h"
#include "StringScanner.h"
#include "String/Encoding.h"
#undef _SCL_SECURE_NO_WARNINGS
#define _SCL_SECURE_NO_WARNINGS
#pragma warning(push, 0)
# include <boost/xpressive/xpressive.hpp>
#pragma warning(pop)
namespace Themis {

CStringScanner::CStringScanner(const string& str) : m_sz(str), m_nPos(0) {
}

int CStringScanner::ToNext() {
    if (IsMatch(L'\r', L'\n')) {
     // m_nLine++;  //s[
        return 2;
    } else if (IsMatch(L'\n')) {
     // m_nLine++;  //s[
        return 1;
    } else {
        m_nPos++;
        return 1;
    }
}

int CStringScanner::Copy1(string& str) {
    if (IsMatch(L'\r')) {
        if (IsMatch(L'\n')) {   //CR+LF
            str += L"\r\n";
         // m_nLine++;  //s[
            return 2;
        } else {
            str += L"\r";
            return 1;
        }
    } else if (IsMatch(L'\n')) {
        str += string(Get1(), 1);
     // m_nLine++;  //s[
        return 1;
    } else {
        Copy1Byte(str);
        return 1;
    }
}

bool CStringScanner::IsMatchRegex(const string& str) {
    const std::wstring chars = m_sz.ToCharArray();
    boost::xpressive::wsregex regex = boost::xpressive::wsregex::compile(str.ToCharPtr());
    boost::xpressive::wsmatch m;
    if (boost::xpressive::regex_match(chars.begin() + m_nPos, chars.end(), m, regex,
        boost::xpressive::regex_constants::match_continuous)) { //[ɂ̂݃}b`
        int n = m.length();
        while (0 < n) {
            n -= ToNext();
        }
        Themis_DebugAssert(n == 0);
        return true;
    }
    return false;
}

int CStringScanner::GetDigit() {
    //[whitespace] [sign] digits
    while (IsBlank()) {
        AddPos(1);
    }
    wchar_t cSign = L'+';
    if (IsMatch(L'-') || IsMatch(L'+')) {
        AddPos(-1);
        cSign = Get1();
        AddPos(1);
    }
    int n = 0;
    while (IsDigit()) {
        n *= 10;
        n += Get1() - L'0';
        AddPos(1);
    }
    return cSign==L'-'? -n: n;
}

int CStringScanner::GetHex() {
    //{digits|a|b|c|d|e|f}
    int n = 0;
    while (*this) {
        if (IsDigit()) {
            n *= 16;
            n += Get1() - L'0';
            AddPos(1);
        } else if (L'a' <= Get1() && Get1() <= L'f') {
            n *= 16;
            n += Get1() - L'a' + 0x0a;
            AddPos(1);
        } else if (L'A' <= Get1() && Get1() <= L'F') {
            n *= 16;
            n += Get1() - L'A' + 0x0a;
            AddPos(1);
        } else {
            break;
        }
    }
    return n;
}

int CStringScanner::GetBin() {
    //{0|1}
    int n = 0;
    while (IsBin()) {
        n *= 2;
        n += Get1() - L'0';
        AddPos(1);
    }
    return n;
}

double CStringScanner::GetFloat() {
    //[whitespace] [sign] [digits] [.digits] [{d|D|e|E} [sign] digits]
    double d = atof(Encoding::GetDefault().GetBytes(Substring(GetPos())).c_str());
    //FXlĂ݂ǁAxƂʓ|Ȃ̂atof()ĝԂ悳B
    while (IsBlank()) { AddPos(1); }
    if (IsMatch(L'-') || IsMatch(L'+')) {}
    while (IsDigit()) { AddPos(1); }
    if (IsMatch(L'.')) {
        while (IsDigit()) { AddPos(1); }
        if (IsMatch(L'd') || IsMatch(L'D') || IsMatch(L'e') || IsMatch(L'E')) {
            if (IsMatch(L'-') || IsMatch(L'+')) {}
            while (IsDigit()) { AddPos(1); }
        }
    }
    return d;
}

string CStringScanner::GetLabel() {
    string str;
    if (IsCSymbolFirst()) {
        do {
            str += string(Get1(), 1);
            AddPos(1);
        } while (IsCSymbol());
    }
    return str;
}

int CStringScanner::SkipSpace() {
    int nLine = 0;
    while (true) {
        if (IsBlank()) {
            AddPos(1);
        } else if (IsMatch(L'\r', L'\n')) {
            AddPos(2);
            nLine++;
        } else if (IsMatch(L'\n')) {
            AddPos(1);
            nLine++;
        } else {
            break;
        }
    }
    return nLine;
}

}
