#include "stdafx.h"
#include <Themis/Windows/System.h>
#include <Themis/Windows/Process.h>
#include <Themis/Windows/MessageBox.h>
#include <Themis/AppFrameBase.h>
#include <Themis/CommandLine.h>
#include <Themis/Allocator.h>
#include <Themis/Window.h>
#include <Themis/TrayIcon.h>
#include <Themis/Graphics.h>
#include <Themis/Image.h>
#include <Themis/Layout.h>
#include <Themis/Config.h>
#include <Themis/File.h>
#include <Themis/CommDlg.h>
#include <Themis/Display.h>
#include <Themis/ProcessExclusion.h>
#include <Themis/Rand.h>
#include <Themis/StringScanner.h>
#include <Themis/DirObserver.h>
#include <Themis/using.h>
#include "resource.h"
#include "Dialogs.h"


const int c_nBuildCount = 15;


class CAppWindow: private IMsgHooker {
    static const UINT_PTR   c_nTimerId = 100;
    const bool              m_bOnce;
    auto_ptr<CWindow>       m_pWindow;
    auto_ptr<CTrayIcon>     m_pTrayIcon;
    auto_ptr<CDirObserver>  m_pDirObserver;
    vector<CFileName>       m_vctFiles;
    int                     m_nCurrent;
    int                     m_nTimerSec;
    bool                    m_bSuspended;
    CFileName               m_oCurrentFile;
public:
    CAppWindow(bool b): m_bOnce(b) {
        m_pWindow.reset(new CWindow);
        m_pWindow->GetHook().Add(this);
        m_pDirObserver.reset(new CDirObserver);
    }

    ~CAppWindow() {
    }

    int Run(CAppFrameBase& rAppFrame) {
        CWindowParameter oPrm;
        oPrm.GetStyle() = WS_POPUP;
        oPrm.GetExStyle() = 0;
        m_pWindow->Create(oPrm);
     // m_pWindow->Show(rAppFrame.GetCmdShow());
        m_pWindow->WaitForQuit();
        return 0;
    }

private:
    bool HookMsg(CMsgHookPrms& rPrms) {
        if (rPrms.GetMsg() == WM_TIMER) {
            ChangeAll();
            return true;
        } else if (rPrms.GetMsg() == WM_COMMAND) {
            switch (rPrms.GetWPrmLoWord()) {
            case ID__40005: // ǎύX
                ChangeAll();
                break;
            case ID__40007: // ݂̕ǎ̌̃t@CJ
                Process::Start(m_oCurrentFile.Get());
                break;
            case ID__40002: // ꎞ~
                {
                    CConfig oConfig;
                    LoadConfig(oConfig);
                    oConfig.Bool(L"Suspend", true);
                    oConfig.Save();
                    ChangeIcon(oConfig);
                    if (oConfig.Bool(L"RemoveOnSuspend")) { // ꎞ~ɕǎO
                        CSystem::UnsetWallpaper();
                    }
                }
                break;
            case ID__40003: // ĊJ
                {
                    CConfig oConfig;
                    LoadConfig(oConfig);
                    oConfig.Bool(L"Suspend", false);
                    oConfig.Save();
                    if (ChangeIcon(oConfig)) {
                        ChangeTimer(oConfig);
                        ChangeWallpaper(oConfig);
                    }
                }
                break;
            case ID__40001: // tH_I
                {
                    CConfig oConfig;
                    LoadConfig(oConfig);
                    CChooseDir oDlg;
                    oDlg.SetOwner(m_pWindow->GetHandle());
                    oDlg.SetTitle(L"摜tH_̐ݒ");
                    oDlg.SetDefault(GetImageDir(oConfig));
                    if (oDlg.Show()) {
                        oConfig.File(L"ImageDir", oDlg.GetResult().Get());
                        oConfig.Save();
                    }
                }
                break;
            case ID__40008: // ݒ
                {
                    CConfig oConfig;
                    LoadConfig(oConfig);
                    CConfigDialog(oConfig).Open(rPrms.GetWnd());
                    oConfig.Save();
                }
                break;
            case ID__40009: // st@ĈtH_J
                Process::Start(CFileNameModuleDir().Get());
                break;
            case ID__40010: // zzTCgJ
                Process::Start(L"http://hp.vector.co.jp/authors/VA039571/");
                break;
            case ID__40011: // ҂փ[𑗂
                Process::Start(L"http://tqzh.tk:24497/site/mail/");
                break;
            case ID__40004: // I
                m_pWindow->Close();
                break;
            case ID__40006: // RtBOύXʒmipj
                {
                    CConfig oConfig;
                    LoadConfig(oConfig);
                    ChangeIcon(oConfig);
                    ChangeTimer(oConfig);
                    ChangeWallpaper(oConfig);
                }
                break;
            default:
                return false;
            }
            return true;
        } else if (rPrms.GetMsg() == WM_APP + 1) { // }EXbZ[W
            UINT nMouseMsg = m_pTrayIcon->GetMouseMsg(rPrms);
            if (nMouseMsg == WM_RBUTTONUP) {
                m_pWindow->ShowPopupMenu(IDR_MENU1);
                return true;
            } else if (nMouseMsg == WM_LBUTTONDBLCLK) {
                CConfig oConfig;
                LoadConfig(oConfig);
                switch (oConfig.Num(L"DoubleClick")) {
                case 0: m_pWindow->PostMsg(WM_COMMAND, ID__40001, 0); break; //tH_I
                case 1: m_pWindow->PostMsg(WM_COMMAND, ID__40005, 0); break; //ǎύX
                case 2: m_pWindow->PostMsg(WM_COMMAND, ID__40007, 0); break; //݂̕ǎ̌̃t@CJ
                case 3:
                    if (oConfig.Bool(L"Suspend")) {
                        m_pWindow->PostMsg(WM_COMMAND, ID__40003, 0);   //ĊJ
                    } else {
                        m_pWindow->PostMsg(WM_COMMAND, ID__40002, 0);   //ꎞ~
                    }
                    break;
                case 4: m_pWindow->PostMsg(WM_COMMAND, ID__40008, 0); break; //ݒ
                }
                return true;
            } else {
                return false;
            }
        } else if (m_pTrayIcon.get() != NULL && m_pTrayIcon->OnTaskbarCreate(rPrms)) { // ACRĕ\
            return true;
        } else if (rPrms.GetMsg() == WM_CREATE) {
            CConfig oConfig;
            LoadConfig(oConfig);
            // 摜ǂݍ݂̐ݒ
            CImageLoader::GetFactoryList().Add(new CSpiLoaderFactory(CFileNameModuleDir(oConfig.File(L"SpiDir"))));
            CImageLoader::GetFactoryList().Add(new CJpgLoaderFactory);
            CImageLoader::GetFactoryList().Add(new CPngLoaderFactory);
            CImageLoader::GetFactoryList().Add(new CBmpLoaderFactory);
            CImageLoader::GetFactoryList().Add(new COleLoaderFactory);
            m_pTrayIcon.reset(new CTrayIcon(m_pWindow->GetHandle(), WM_APP + 1));
            if (m_bOnce) {
                ChangeWallpaper(oConfig);
                m_pWindow->PostClose();
            } else {
                // ACR̐ݒ
                m_bSuspended = oConfig.Bool(L"Suspend");
                m_pTrayIcon->SetIcon(m_bSuspended? IDI_ICON2: IDI_ICON1);
                m_pTrayIcon->SetTip(string::Format(L"񂩂%s build-%d", m_bSuspended? L" [ꎞ~]": L"", c_nBuildCount));
                // ^C}[ݒ
                m_nTimerSec = oConfig.Num(L"Interval(sec)");
                m_pWindow->SetTimer(c_nTimerId, m_nTimerSec * 1000);
                // ǎύX
                if (!oConfig.Bool(L"WaitFirst")) {
                    ChangeWallpaper(oConfig);
                }
            }
            return true;
        } else if (rPrms.GetMsg() == WM_CLOSE) {
            m_pWindow->Delete();
            return true;
        } else if (rPrms.GetMsg() == WM_DESTROY) {
            m_pTrayIcon->UnsetIcon();
            return true;
        } else {
            return false;
        }
    }

    void ChangeAll() {
        CConfig oConfig;
        LoadConfig(oConfig);
        ChangeIcon(oConfig);
        ChangeTimer(oConfig);
        ChangeWallpaper(oConfig);
    }

    bool ChangeIcon(const CConfig& oConfig) {
        if (m_bSuspended != oConfig.Bool(L"Suspend")) {
            m_bSuspended = oConfig.Bool(L"Suspend");
            m_pTrayIcon->SetIcon(m_bSuspended? IDI_ICON2: IDI_ICON1);
            m_pTrayIcon->SetTip(string::Format(L"񂩂%s build-%d", m_bSuspended? L" [ꎞ~]": L"", c_nBuildCount));
            return true;
        }
        return false;
    }

    void ChangeTimer(const CConfig& oConfig) {
        if (m_nTimerSec != oConfig.Num(L"Interval(sec)")) {
            m_nTimerSec = oConfig.Num(L"Interval(sec)");
            m_pWindow->UnsetTimer(c_nTimerId);
            m_pWindow->SetTimer(c_nTimerId, m_nTimerSec * 1000);
        }
    }

    void ChangeWallpaper(const CConfig& oConfig) {
        if (!oConfig.Bool(L"Suspend")) {
            // t@CXg̍쐬
            CFileName imageDir = GetImageDir(oConfig);
            if (!CFile(imageDir).ExistsDir()) {
                MessageBox::Show(m_pWindow->GetHandle(),
                    L"摜tH_ '" +  imageDir.Get() + L"' ܂",
                    L"G[",
                    MessageBoxButtons::OK, MessageBoxIcon::Error);
                return;
            }
            if (m_pDirObserver->GetDirName() != imageDir) {
                m_pDirObserver->Begin(imageDir);
                m_vctFiles.clear();
            }
            if (m_vctFiles.empty() || m_pDirObserver->IsChanged()) {
                m_pDirObserver->ResetChangeCount();
                m_vctFiles = CFile::Find(m_pDirObserver->GetDirName(), -1);
                std::random_shuffle(m_vctFiles.begin(), m_vctFiles.end(), CRand::GetInstance());
            }
            //ǎɂt@C
            if (m_vctFiles.empty()) {
                return;
            }
            m_nCurrent = int((m_nCurrent + 1) % m_vctFiles.size());
            m_oCurrentFile = m_vctFiles[m_nCurrent];

            //ǂݍŏŐݒ
            try {
                CFileNameModuleDir oFileName(oConfig.File(L"SaveFile"));
                std::tr1::shared_ptr<CPlaneEx> pPlane(new CDib32);
                std::tr1::shared_ptr<CPlaneEx> pTempPlane(new CMemoryPlane32);
                CImageLoader::Load(*pPlane, new CExistFileData(m_oCurrentFile));
                if (oConfig.Bool(L"Resize")) { // TCY
                    CRect src, dst;
                    vector<CRect> bgRects;
                    COLORREF bgColor = CLR_INVALID;
                    if (oConfig.Bool(L"ResizeFixedAspectRatio")) {
                        AutoStretchingLayout layout;
                        layout.SetScreen(CDisplay::GetCurrentMode().oSize.ToRect());
                        layout.SetDocument(pPlane->GetSize().ToRect());
                        layout.SetReduction(true);
                        layout.SetMagnification(true);
                        layout.SetStickOut(false);
                        layout.SetAlignH(oConfig.Num(L"ResizeAlignH"));
                        layout.SetAlignV(oConfig.Num(L"ResizeAlignV"));
                        src = layout.GetSrc();
                        dst = layout.GetDst();
                        bgRects = layout.GetScreen().GetBlanks(dst);
                        switch (oConfig.Num(L"ResizeBgColor")) {
                        default:
                        case 0:
                            if (layout.IsStretchToHorz()) {
                                switch (layout.GetAlignH()) {
                                case 0: bgColor = pPlane->GetAverageColor(CRect(src.right - 2, src.top, 2, src.GetHeight())); break;
                                case 1:
                                    {
                                        COLORREF c1 = pPlane->GetAverageColor(CRect(src.left, src.top, 2, src.GetHeight()));
                                        COLORREF c2 = pPlane->GetAverageColor(CRect(src.right - 2, src.top, 2, src.GetHeight()));
                                        bgColor = RGB(
                                            (GetRValue(c1) + GetRValue(c2)) / 2, 
                                            (GetGValue(c1) + GetGValue(c2)) / 2, 
                                            (GetBValue(c1) + GetBValue(c2)) / 2);
                                    }
                                    break;
                                }
                                case 2: bgColor = pPlane->GetAverageColor(CRect(src.left, src.top, 2, src.GetHeight())); break;
                            } else {
                                switch (layout.GetAlignV()) {
                                case 0: bgColor = pPlane->GetAverageColor(CRect(src.right, src.top - 2, src.GetWidth(), 2)); break;
                                case 1:
                                    {
                                        COLORREF c1 = pPlane->GetAverageColor(CRect(src.left, src.top, src.GetWidth(), 2));
                                        COLORREF c2 = pPlane->GetAverageColor(CRect(src.right, src.top - 2, src.GetWidth(), 2));
                                        bgColor = RGB(
                                            (GetRValue(c1) + GetRValue(c2)) / 2, 
                                            (GetGValue(c1) + GetGValue(c2)) / 2, 
                                            (GetBValue(c1) + GetBValue(c2)) / 2);
                                    }
                                    break;
                                case 2: bgColor = pPlane->GetAverageColor(CRect(src.left, src.top, src.GetWidth(), 2)); break;
                                }
                            }
                            if (bgColor == CLR_INVALID) {
                                bgColor = pPlane->GetAverageColor(pPlane->GetSize().ToRect());
                            }
                            break;
                        case 1:
                            bgColor = ::GetSysColor(COLOR_DESKTOP);
                            break;
                        }
                    } else {
                        src = pPlane->GetSize().ToRect();
                        dst = CDisplay::GetCurrentMode().oSize.ToRect();
                    }   
                    if (oConfig.Bool(L"ClearType")) {
                        pPlane.swap(pTempPlane);
                        pPlane->SetSize(CPlaneEx::GetClearTypeSize(dst.GetSize()));
                        pPlane->StretchBltLanczos3(*pTempPlane, src, pTempPlane->GetSize().ToRect());
                        pPlane.swap(pTempPlane);
                        pPlane->SetSize(CDisplay::GetCurrentMode().oSize);
                        pPlane->ClearTypeBlt(*pTempPlane, dst.GetPoint(), pTempPlane->GetSize().ToRect());
                    } else {
                        pPlane.swap(pTempPlane);
                        pPlane->SetSize(CDisplay::GetCurrentMode().oSize);
                        pPlane->StretchBltLanczos3(*pTempPlane, dst, src);
                    }
                    if (bgColor != CLR_INVALID) {
                        for (size_t i = 0, n = bgRects.size(); i < n; i++) {
                            pPlane->Fill(bgColor, bgRects[i]);
                        }
                    }
                }
                if (oConfig.Bool(L"Sharp")) { // V[v
                    pPlane.swap(pTempPlane);
                    pPlane->SetSize(pTempPlane->GetSize());
                    pPlane->SharpBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
                }
                CBmpSaver().Save(oFileName, *pPlane, pPlane->GetSize().ToRect());
                CSystem::SetWallpaper(oFileName.Get());
            } catch (std::exception&) {
                Themis_DebugWarning(L"Ȃ񂩃G[ۂ");
            }
        }
    }

    CFileName GetImageDir(const CConfig& oConfig) const {
        string str = oConfig.File(L"ImageDir");
        CFileNameModuleDir dir(CFileName::AddDirSeparator(str));
        return dir;
    }

    void LoadConfig(CConfig& oConfig) {
        oConfig.File(L"ImageDir", L"./image/");
        oConfig.File(L"SaveFile", L"%windir%/Web/Wallpaper/񂩂.bmp");
        oConfig.File(L"SpiDir", CSpi::GetInstallDir());
        if (oConfig.File(L"SpiDir").GetLength() <= 0) {
            oConfig.File(L"SpiDir", L"./spi/");
        }
        oConfig.Num(L"Interval(sec)", 30 * 60); //30
        oConfig.Bool(L"WaitFirst", false);
        oConfig.Bool(L"Suspend", false);
        oConfig.Bool(L"Resize", true);
        oConfig.Bool(L"ResizeFixedAspectRatio", false);
        oConfig.Num(L"ResizeAlignH", 1);
        oConfig.Num(L"ResizeAlignV", 0);
        oConfig.Num(L"ResizeBgColor", 1);
        oConfig.Bool(L"ClearType", false);
        oConfig.Bool(L"Sharp", false);
        oConfig.Bool(L"RemoveOnSuspend", false);
        oConfig.Num(L"DoubleClick", 2);
        oConfig.ChanegName(L"Cleartype", L"ClearType");
        CProcessExclusionLock oLock(CSystem::GetModuleFileName() + L".cfg");
        if (!oConfig.Load()) {
            oConfig.Save(); //݂ȂꍇftHgݒo
        }
    }
};

class CAppFrame : public CSingleAppFrameBase {
    int InnerRun() {
        bool bNoRun = false;
        bool bOnce = false;
        if (GetCmdLine() != NULL && GetCmdLine()[0] != L'\0') {
            CProcessExclusionLock oLock(CSystem::GetModuleFileName() + L".cfg");
            CCommandLine oCmdLine(GetCmdLine());
            for (int i = 0, n = oCmdLine.GetCount(); i < n; i++) {
                if (oCmdLine.IsMatch(i, L"/suspend")) {
                    bNoRun = true;
                    CConfig oConfig;
                    oConfig.Load();
                    if (!oConfig.Bool(L"Suspend") && oConfig.Bool(L"RemoveOnSuspend")) {// ꎞ~ɕǎO
                        CSystem::UnsetWallpaper();
                    }
                    oConfig.Bool(L"Suspend", true);
                    oConfig.Save();
                } else if (oCmdLine.IsMatch(i, L"/resume")) {
                    bNoRun = true;
                    CConfig oConfig;
                    oConfig.Load();
                    oConfig.Bool(L"Suspend", false);
                    oConfig.Save();
                } else if (oCmdLine.IsMatch(i, L"/once")) {
                    bOnce = true;
                }
            }
        }

        if (!bNoRun && IsFirstRun()) {
            return CAppWindow(bOnce).Run(*this);
        } else {
            CWnd oWindow = CSystem::GetWindow(CSystem::GetModuleFileName());
            if (oWindow.Exists()) {
                oWindow.PostMsg(WM_COMMAND, ID__40006, 0); //XVʒmbZ[W(p)
            }
            return 0;
        }
    }
};

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow) {
    CAppFrame theApp;
    theApp.Initialize(hInstance, hPrevInstance, lpCmdLine, nCmdShow);
    return theApp.Run();
}

